#ifndef __name_table_ext__
#define __name_table_ext__

//------------------------------------------------------------------------------
// Файл name_table_ext.h - содержит описания типов и данных, 
// используемых функциями, работающими с таблицей имен
//------------------------------------------------------------------------------

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <map>
#include <vector>

using namespace std;

#include "app_types.h"

// Упрощенное переопределение таблицы имен
typedef map<string, application*, less<string> > NameTable;

// Ссылка на таблицу имен, заданную в виде глобальной пременной,
// так как она является единственной и внешней в реализуемой
// семантической модели программы
extern NameTable table;

// Адрес памяти, используемой для размещения переменных
extern int glob_addr;

// Максимальная длина вектора
int const maxVecLen = 32567;

// Вывод таблицы имен в поток, закрепленный за файлом или на экран
// в зависимости от указания потока в качестве второго аргумента.
// По умолчанию осуществляется вывод в cout.
void Out(NameTable &table, FILE *os = stdout);

// Проверка всех меток на наличие объявлений
bool AllLabelsDefined(NameTable &table);

// Очистка таблицы имен по окончании работы от назначений
void Clear(NameTable &table);

#endif
