//------------------------------------------------------------------------------
// parser.cpp - функции, осуществляющие распознавание
//------------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "scanner_ext.h"
#include "name_table_ext.h"

//------------------------------------------------------------------------------

void er(int);

//------------------------------------------------------------------------------

bool Declaration();
bool Statement();
bool Etc_stmts();
bool No_lab_stmt();
bool Ass_stmt_continue_1();
bool Ass_stmt_continue_2();
bool Assign_stmt();
bool Variable();
bool Expression();
bool Addition();
bool Product();
bool Term();
bool Number();
bool Case_stmt();
bool Loop_stmt();
bool Guard_set();
bool Guard();
bool Read_stmt();
bool Write_stmt();
bool Specificator();
bool Goto_stmt();
bool Abort_stmt();

//------------------------------------------------------------------------------
bool Program() {
//_0:
    if(lc==kwBegin) {nxl(); goto _1;}
    return false;
  _1:
    if(Declaration()) {goto _2;}
    if(Statement()) {goto _2;}
    if(lc==lexSemicolon) {nxl(); goto _1;}
    if(lc==kwEnd) {nxl(); goto _end;}
    er(7); return false;
  _2:
    if(lc==lexSemicolon) {nxl(); goto _1;}
    if(lc==kwEnd) {nxl(); goto _end;}
    er(7); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Declaration() {
	// Дополнительный вектор позволяет созранить указатели на переменные,
	// чтобы потом добавить ко всем недостающий теип и вычислить адреса.
	vector<app_variable*> tmp_var_vec;
	// Указатель на текущую переменную
	app_variable *tmp_app_var;
	// Временная переменная для сохранения текущего типа
	scalType tmp_type;
	// Переменная для размера текущего типа
	int type_size;
//_0:
    if(lc==kwVar) {nxl(); goto _1;}
    return false;
  _1:
    if(lc==lexId) 
	{
		NameTable::iterator i = table.find(lv);
		if(i == table.end()) // Элемент отсутствовал
		{
			// Необходимо добавить переменную в таблицк имен

			// В начале создается элемент таблицы имен
			tmp_app_var = new app_variable;
			// Пока длина = 0 (скаляр)
			tmp_app_var->_len = 0;
			table[lv] = tmp_app_var;
			// Ссылка добавляется в вектор временных переменных
			tmp_var_vec.push_back(tmp_app_var);
		}
		else // Элемент уже присутствует
		{
			// Необходимо выдать ошибку: 
			// повторное объявление переменной или метки
			er(25); return false;
		}
		nxl(); goto _2;
	}
    er(8); return false;
  _2:
    if(lc==lexLftSqBr) {nxl(); goto _3;}
    if(lc==lexComma) {nxl(); goto _1;}
    if(lc==lexColon) {nxl(); goto _6;}
    er(8); return false;
  _3:
    if(lc==lexInt) 
	{
		// Пришел размер вектора. 
		// Необходимо проверить его значение
		if(unum > 0 && unum < maxVecLen)
		{
			// Осуществляем установку длины
			tmp_app_var->_len = unum;
		}
		else
		{
			// Ошибка. Недопустимая длина вектора
			er(34); return false;
		}
		nxl(); goto _4;
	}
    er(8); return false;
  _4:
    if(lc==lexRghSqBr) {nxl(); goto _5;}
    er(8); return false;
  _5:
    if(lc==lexComma) {nxl(); goto _1;}
    if(lc==lexColon) {nxl(); goto _6;}
    er(8); return false;
  _6:
    if(lc==kwInt) 
	{
		tmp_type = INT_TYPE;
		type_size = 4; // 32-х разрядная целочисленная переменная 
		nxl(); goto _end;
	}
    if(lc==kwFloat) 
	{
		tmp_type = FLOAT_TYPE;
		type_size = 8; // 64-х разрядная действительная переменная 
		nxl(); goto _end;
	}
    er(8); return false;
  _end:
	// Окончательное заполнение параметров переменных
	vector<app_variable*>::iterator i;
	for(i = tmp_var_vec.begin(); i != tmp_var_vec.end(); ++i)
	{
		(*i)->_typ = tmp_type;
		(*i)->_addr = glob_addr;
		if((*i)->_len == 0)
			glob_addr += type_size; // для скаляра
		else
			glob_addr += type_size * (*i)->_len; // для вектора
	}
    return true;
}

//------------------------------------------------------------------------------
bool Statement() {
	string id; // буфер для сохранения идентификатора
//_0:
    if(lc==lexId) 
	{
		id = lv; // Сохраняем до опрделения котекста
		// Весь разбор имени осуществляем после перехода на метку _1
		// после опредения символа, следующего за идентификатором.
		nxl(); goto _1;
	}
    if(Etc_stmts()) {goto _end;}
    return false;
  _1:
    if(lc==lexColon) // Обнаружена новая метка.
	{
		// Это - анализ имени метки
		NameTable::iterator i = table.find(id);
		if(i != table.end()) 
		{ 
			// Имя уже присутствует в таблице. Надо выяснить, является оно
			// меткой, встретившейся ранее в goto, или уже существующей меткой
			// или именем переменной. Приходится использовать RTTI
			// Следует динамическое приведение данных
			if(app_variable* p_var = dynamic_cast<app_variable*>((*i).second))
			{
				// Обнаружена переменная с таким же именем, что и у метки. 
				// Это является ошибкой.
				er(26); return false;
			}
			else if(app_label* p_lab = dynamic_cast<app_label*>((*i).second))
			{
				// Обнаружена метка. Надо проверить, была ли она объявлена.
				if(p_lab->wasDcl)
				{
					// Это является ошибкой.
					er(26); return false;
				}
				else
				{
					// Метка раньше встречалась только в goto. Можно объявлять
					p_lab->wasDcl = true;
				}
			}
		}
		else
		{
			// Надо записать метку в таблицу, зарегистрировав ее объявление.
			// В начале создается элемент таблицы имен
			app_label *tmp_app_lab = new app_label(true);
			table[id] = tmp_app_lab;
		}
		nxl(); goto _2;
	}
    if(lc==lexLftSqBr) 
	{
		// Это - анализ имени переменной.
		NameTable::iterator i = table.find(id);
		if(i != table.end()) 
		{ 
			// Имя уже присутствует в таблице. Надо выяснить, является оно
			// именем переменной. Приходится использовать RTTI
			// Следует динамическое приведение данных
			if(app_variable* p_var = dynamic_cast<app_variable*>((*i).second))
			{
				// Переменная с таким же именем уже объявлена
				// Все корректно. Можно с ней работать.

				// Контекстный анализ переменной на принадлежность к векторам
				if(p_var->_len == 0)
				{
					// Это скаляр, что недопустимо в контексте нашего языка.
					// В других, может быть, и позволено. Но не здесь!
					er(32); return false;
				}
				;;;
			}
			else 
			{
				// Объявлена, но не переменная. Ошибка
				er(27); return false;
			}
		}
		else
		{
			// Переменная оказалась предварительно не объявленной. Ошибка.
			er(28); return false;
		}
		nxl(); goto _3;
	}
    if(lc==lexComma) 
	{
		// Это - тоже анализ имени переменной.
		NameTable::iterator i = table.find(id);
		if(i != table.end()) 
		{ 
			// Имя уже присутствует в таблице. Надо выяснить, является оно
			// именем переменной. Приходится использовать RTTI
			// Следует динамическое приведение данных
			if(app_variable* p_var = dynamic_cast<app_variable*>((*i).second))
			{
				// Переменная с таким же именем уже объявлена
				// Все корректно. Можно с ней работать.

				// Контекстный анализ переменной на принадлежность к скалярам
				if(p_var->_len > 0)
				{
					// Это вектор, что недопустимо в контексте нашего языка.
					// В других, может быть, и позволено. Но не здесь!
					er(33); return false;
				}
				;;;
			}
			else 
			{
				// Объявлена, но не переменная. Ошибка
				er(27); return false;
			}
		}
		else
		{
			// Переменная оказалась предварительно не объявленной. Ошибка.
			er(28); return false;
		}
		nxl(); goto _4;
	}
    if(lc==lexAssign) 
	{
		// А это - тоже анализ имени переменной.
		NameTable::iterator i = table.find(id);
		if(i != table.end()) 
		{ 
			// Имя уже присутствует в таблице. Надо выяснить, является оно
			// именем переменной. Приходится использовать RTTI
			// Следует динамическое приведение данных
			if(app_variable* p_var = dynamic_cast<app_variable*>((*i).second))
			{
				// Переменная с таким же именем уже объявлена
				// Все корректно. Можно с ней работать.

				// Контекстный анализ переменной на принадлежность к скалярам
				if(p_var->_len > 0)
				{
					// Это вектор, что недопустимо в контексте нашего языка.
					// В других, может быть, и позволено. Но не здесь!
					er(33); return false;
				}
				;;;
			}
			else 
			{
				// Объявлена, но не переменная. Ошибка
				er(27); return false;
			}
		}
		else
		{
			// Переменная оказалась предварительно не объявленной. Ошибка.
			er(28); return false;
		}
		nxl(); goto _5;
	}
    er(9); return false;
  _2:
    if(No_lab_stmt()) {goto _end;}
	goto _end;
  _3:
    if(Ass_stmt_continue_1()) {goto _end;}
    er(9); return false;
  _4:
    if(Ass_stmt_continue_2()) {goto _end;}
    er(9); return false;
  _5:
    if(Expression()) {goto _end;}
    er(9); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool No_lab_stmt() {
//_0:
    if(lc==kwSkip) {nxl(); goto _end;}
    if(Assign_stmt()) {goto _end;}
    if(Case_stmt())   {goto _end;}
    if(Loop_stmt())   {goto _end;}
    if(Read_stmt())   {goto _end;}
    if(Write_stmt())  {goto _end;}
    if(Goto_stmt())   {goto _end;}
    if(Abort_stmt())  {goto _end;}
    return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Etc_stmts() {
//_0:
    if(lc==kwSkip) {nxl(); goto _end;}
    if(Case_stmt())   {goto _end;}
    if(Loop_stmt())   {goto _end;}
    if(Read_stmt())   {goto _end;}
    if(Write_stmt())  {goto _end;}
    if(Goto_stmt())   {goto _end;}
    if(Abort_stmt())  {goto _end;}
    return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Ass_stmt_continue_1() {
//_0:
    if(Expression()) {goto _1;}
    return false;
  _1:
    if(lc==lexRghSqBr) {nxl(); goto _2;}
    er(10); return false;
  _2:
    if(lc==lexComma) {nxl(); goto _3;}
    if(lc==lexAssign) {nxl(); goto _4;}
    er(10); return false;
  _3:
    if(Ass_stmt_continue_2()) {goto _end;}
    er(10); return false;
  _4:
    if(Expression()) {goto _end;}
    er(10); return false;  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Ass_stmt_continue_2() {
//_0:
    if(Assign_stmt()) {goto _1;}
    return false;
  _1:
    if(lc==lexComma) {nxl(); goto _2;}
    er(11); return false;
  _2:
    if(Expression()) {goto _end;}
    er(11); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Assign_stmt() {
//_0:
    if(Variable()) {goto _1;}
    return false;
  _1:
    if(lc==lexAssign) {nxl(); goto _4;}
    if(lc==lexComma)  {nxl(); goto _2;}
    er(12); return false;
  _2:
    if(Assign_stmt()) {goto _3;}
    er(12); return false;
  _3:
    if(lc==lexComma) {nxl(); goto _4;}
    er(12); return false;
  _4:
    if(Expression()) {goto _end;}
    er(12); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Variable() {
	app_variable* p_var; // Указатель на текущую переменную
//_0:
    if(lc==lexId) 
	{
		// Это - анализ имени переменной.
		NameTable::iterator i = table.find(lv);
		if(i != table.end()) 
		{ 
			// Имя уже присутствует в таблице. Надо выяснить, является оно
			// именем переменной. Приходится использовать RTTI
			// Следует динамическое приведение данных
			if(p_var = dynamic_cast<app_variable*>((*i).second))
			{
				// Переменная с таким же именем уже объявлена
				// Все корректно. Можно с ней работать.
				;;;
			}
			else 
			{
				// Объявлена, но не переменная. Ошибка
				er(27); return false;
			}
		}
		else
		{
			// Переменная оказалась предварительно не объявленной. Ошибка.
			er(30); return false;
		}
		nxl(); goto _1;
	}
    return false;
  _1:
    if(lc==lexLftSqBr) 
	{
		// Контекстный анализ переменной на принадлежность к векторам
		if(p_var->_len == 0)
		{
			// Это скаляр, что недопустимо в контексте нашего языка.
			// В других, может быть, и позволено. Но не здесь!
			er(32); return false;
		}
		nxl(); goto _2;
	}
	// Контекстный анализ переменной на принадлежность к скалярам
	if(p_var->_len > 0)
	{
		// Это вектор, что недопустимо в контексте нашего языка.
		// В других, может быть, и позволено. Но не здесь!
		er(33); return false;
	}
    goto _end;
  _2:
    if(Expression()) {goto _3;}
    er(13); return false;
  _3:
    if(lc==lexRghSqBr) {nxl(); goto _end;}
    er(13); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Expression() {
//_0:
    if(Addition()) {goto _1;}
    return false;
  _1:
    if(lc==lexEQ) {nxl(); goto _2;}
    if(lc==lexGT) {nxl(); goto _2;}
    if(lc==lexLT) {nxl(); goto _2;}
    if(lc==lexNE) {nxl(); goto _2;}
    if(lc==lexGE) {nxl(); goto _2;}
    if(lc==lexLE) {nxl(); goto _2;}
    goto _end;
  _2:
    if(Addition()) {goto _1;}
    er(14); return false;
  _end:
    return true;
}


//------------------------------------------------------------------------------
bool Addition() {
//_0:
    if(Product()) {goto _1;}
    return 0;
  _1:
    if(lc==lexPlus)  {nxl(); goto _2;}
    if(lc==lexMinus) {nxl(); goto _2;}
    goto _end;
  _2:
    if(Product()) {goto _1;}
    er(15); return false;
  _end:
    return true;

}

//------------------------------------------------------------------------------
bool Product() {
//_0:
    if(Term()) {goto _1;}
    return 0;
  _1:
    if(lc==lexStar) {nxl(); goto _2;}
    if(lc==lexSlash) {nxl(); goto _2;}
    if(lc==lexPercent) {nxl(); goto _2;}
    goto _end;
  _2:
    if(Term()) {goto _1;}
    er(16); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Term() {
//_0:
    if(lc==lexMinus)    {nxl(); goto _1;}
    if(lc==lexLftRndBr) {nxl(); goto _2;}
    if(Number())   {goto _end;}
    if(Variable()) {goto _end;}
    return false;
  _1:
    if(lc==lexLftRndBr) {nxl(); goto _2;}
    if(Number())   {goto _end;}
    if(Variable()) {goto _end;}
    er(17); return false;
  _2:
    if(Expression()) {goto _3;}
    er(17); return false;
  _3:
    if(lc==lexRghRndBr) {nxl(); goto _end;}
    er(17); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Number() {
//_0:
    if(lc==lexInt)   {nxl(); goto _end;}
    if(lc==lexFloat) {nxl(); goto _end;}
    return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Case_stmt() {
//_0:
    if(lc==kwCase) {nxl(); goto _1;}
    return false;
  _1:
    if(lc==kwEnd) {nxl(); goto _end;}
    if(Guard_set()) {goto _2;}
    er(18); return false;
  _2:
    if(lc==kwEnd) {nxl(); goto _end;}
    er(18); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Loop_stmt() {
//_0:
    if(lc==kwLoop) {nxl(); goto _1;}
    return false;
  _1:
    if(lc==kwEnd) {nxl(); goto _end;}
    if(Guard_set()) {goto _2;}
    er(19); return false;
  _2:
    if(lc==kwEnd) {nxl(); goto _end;}
    er(19); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Guard_set() {
//_0:
    if(Guard()) {goto _1;}
    return false;
  _1:
    if(lc==kwOr) {nxl(); goto _2;}
    goto _end;
  _2:
    if(Guard()) {goto _1;}
    er(20); return false;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Guard() {
//_0:
    if(Expression()) {goto _1;}
    return false;
  _1:
    if(lc==lexArrow) {nxl(); goto _2;}
    er(21); return false;
  _2:
    if(Statement()) {goto _3;}
    if(lc==lexSemicolon) {nxl(); goto _2;}
    goto _end;
  _3:
    if(lc==lexSemicolon) {nxl(); goto _2;}
    goto _end;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Read_stmt() {
//_0:
    if(lc==kwRead) {nxl(); goto _1;}
    return false;
  _1:
    if(Variable()) {goto _2;}
    er(22); return false;
  _2:
    if(lc==lexComma) {nxl(); goto _1;}
    goto _end;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Write_stmt() {
//_0:
    if(lc==kwWrite) {nxl(); goto _1;}
    return false;
  _1:
    if(lc==lexString) {nxl(); goto _2;}
    if(Expression()) {goto _2;}
    if(Specificator()) {goto _2;}
    er(23); return false;
  _2:
    if(lc==lexComma) {nxl(); goto _1;}
    goto _end;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Specificator() {
//0:
    if(lc==kwSpace) {nxl(); goto _1;}
    if(lc==kwTab)   {nxl(); goto _1;}
    if(lc==kwSkip)  {nxl(); goto _1;}
    return false;
  _1:
    if(Expression()) {goto _end;}
    goto _end;
  _end:
    return true;
}

//------------------------------------------------------------------------------
bool Goto_stmt() {
//_0:
    if(lc==kwGoto) {nxl(); goto _1;}
    return false;
  _1:
    if(lc==lexId) 
	{
		// Это - анализ имени метки, следующей после goto
		NameTable::iterator i = table.find(lv);
		if(i != table.end()) 
		{ 
			// Имя метки уже присутствует в таблице.
			if(app_label* p_lab = dynamic_cast<app_label*>((*i).second))
			{
				// Все прекрасно! используем метку в операторе goto.
				;;;
			}
			else 
			{
				// Остался вариант, когда имя не определяет метку.
				er(31); return false;
			}
		}
		else
		{
			// Надо записать метку в таблицу, 
			// зарегистрировав ее появление только в goto.
			// В начале создается элемент таблицы имен
			app_label *tmp_app_lab = new app_label(false);
			table[lv] = tmp_app_lab;
		}
		nxl(); goto _end;
	}
    er(24); return false;
  _end:
    return true;
}


//------------------------------------------------------------------------------
bool Abort_stmt() {
//_0:
    if(lc==kwAbort) {nxl(); goto _1;}
    return false;
  _1:
    if(lc==lexString) {nxl(); goto _end;}
    goto _end;
  _end:
    return true;
}

//------------------------------------------------------------------------------
